from sklearn_pandas import DataFrameMapper
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder
from sklearn2pmml import _is_categorical
from sklearn2pmml.preprocessing import PMMLLabelEncoder

def make_lightgbm_dataframe_mapper(dtypes, missing_value_aware = True):
	"""Construct a DataFrameMapper for feeding complex data into a LGBMModel.

	Parameters:
	----------

	dtypes: iterable of tuples (column, dtype)

	missing_value_aware: boolean
		If true, use missing value aware transformers.

	Returns:
	-------
	Tuple (DataFrameMapper, list of categorical columns indices)

	"""
	features = list()
	categorical_features = list()
	i = 0
	for column, dtype in dtypes.items():
		if _is_categorical(dtype):
			features.append(([column], PMMLLabelEncoder(missing_values = -1) if missing_value_aware else LabelEncoder()))
			categorical_features.append(i)
		else:
			features.append(([column], None))
		i += 1
	return (DataFrameMapper(features), categorical_features)

def make_lightgbm_column_transformer(dtypes, missing_value_aware = True):
	"""Construct a ColumnTransformer for feeding complex data into a LGBMModel.

	Parameters:
	----------

	dtypes: iterable of tuples (column, dtype)

	missing_value_aware: boolean
		If true, use missing value aware transformers.

	Returns:
	-------
	Tuple (ColumnTransformer, list of categorical column indices)

	"""
	transformers = list()
	categorical_features = list()
	i = 0
	for column, dtype in dtypes.items():
		if _is_categorical(dtype):
			transformers.append((str(column), PMMLLabelEncoder(missing_values = -1) if missing_value_aware else OrdinalEncoder(), [column]))
			categorical_features.append(i)
		else:
			transformers.append((str(column), "passthrough", [column]))
		i += 1
	return (ColumnTransformer(transformers, remainder = "drop"), categorical_features)


from scipy.special import expit, softmax
from sklearn.preprocessing import normalize
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn2pmml.util import eval_rows, to_expr_func, Expression

import numpy

class ExpressionRegressor(BaseEstimator, RegressorMixin):

	def __init__(self, expr):
		if not isinstance(expr, Expression):
			raise TypeError()
		self.expr = expr

	def fit(self, X, y):
		return self

	def predict(self, X):
		func = to_expr_func(self.expr)
		return eval_rows(X, func, dtype = float)

class ExpressionClassifier(BaseEstimator, ClassifierMixin):

	def __init__(self, class_exprs, normalization_method):
		if type(class_exprs) is not dict:
			raise TypeError()
		if not class_exprs:
			raise ValueError()
		for k, v in class_exprs.items():
			if k is None:
				raise ValueError()
			if not isinstance(v, Expression):
				raise TypeError()
		self.class_exprs = class_exprs
		normalization_methods = ["logit", "simplemax", "softmax"]
		if normalization_method not in normalization_methods:
			raise ValueError("Normalization method {0} not in {1}".format(normalization_method, normalization_methods))
		self.normalization_method = normalization_method

	def fit(self, X, y):
		self.classes_ = numpy.unique(y)
		n_classes = len(self.classes_)
		classes = set(self.class_exprs.keys())
		if n_classes == 2 and self.normalization_method == "logit":
			if (len(classes) != 1) or (classes.intersection(self.classes_) != classes):
				raise ValueError()
		elif n_classes >= 2 and self.normalization_method in ["simplemax", "softmax"]:
			if (len(classes) != n_classes) or (classes.intersection(self.classes_) != classes):
				raise ValueError()
		else:
			raise ValueError()
		return self

	def decision_function(self, X):
		y = None
		for clazz in self.classes_:
			try:
				class_expr = self.class_exprs[clazz]				
			except KeyError:
				continue
			class_func = to_expr_func(class_expr)
			class_y = eval_rows(X, class_func, dtype = float)
			if y is None:
				y = class_y
			else:
				y = numpy.vstack((y, class_y))
		if self.normalization_method in ["simplemax", "softmax"]:
			y = y.T.reshape(X.shape[0], -1)
		return y

	def predict(self, X):
		y_proba = self.predict_proba(X)
		indices = numpy.argmax(y_proba, axis = 1)
		return numpy.take(self.classes_, indices, axis = 0)

	def predict_proba(self, X):
		y = self.decision_function(X)
		n_classes = len(self.classes_)
		if n_classes == 2 and self.normalization_method == "logit":
			proba = expit(y)
			clazz = next(iter(self.class_exprs.keys()))
			if (self.classes_).tolist().index(clazz) == 1:
				return numpy.vstack((1 - proba, proba)).T
			else:
				return numpy.vstack((proba, 1 - proba)).T
		elif n_classes >= 2 and self.normalization_method == "simplemax":
			return normalize(y, norm = "l1", axis = 1)
		elif n_classes >= 2 and self.normalization_method == "softmax":
			return softmax(y, axis = 1)
		else:
			raise ValueError()



try:
	from h2o.frame import H2OFrame
except ImportError:
	pass
from sklearn.base import BaseEstimator, TransformerMixin

import warnings

class H2OFrameConstructor(BaseEstimator, TransformerMixin):

	def __init__(self, column_names = None, column_types = None):
		self.column_names = column_names
		self.column_types = column_types

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return H2OFrame(python_obj = X, column_names = self.column_names, column_types = self.column_types)

class H2OFrameCreator(H2OFrameConstructor):

	def __init__(self, column_names = None, column_types = None):
		super(H2OFrameCreator, self).__init__(column_names = column_names, column_types = column_types)
		warnings.warn("{} has been renamed to {}, the alias will be removed in the future".format(H2OFrameCreator.__name__, H2OFrameConstructor.__name__), DeprecationWarning)



from pandas import CategoricalDtype
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
try:
	from sklearn.compose import ColumnTransformer
except ImportError:
	class ColumnTransformer:
		pass
try:
	from sklearn.feature_selection._base import SelectorMixin
except ImportError:
	from sklearn.feature_selection.base import SelectorMixin
from sklearn.feature_selection import SelectFromModel
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn_pandas import DataFrameMapper
from sklearn2pmml.resources import _package_classpath
from subprocess import PIPE, Popen
from zipfile import ZipFile

import dill
import joblib
import numpy
import os
import pandas
import platform
import re
import sklearn
import sklearn_pandas
import tempfile
import warnings

from .metadata import __copyright__, __license__, __version__
from .pipeline import PMMLPipeline

def _is_categorical(dtype):
	if dtype == object or dtype == str or dtype == bool:
		return True
	elif _is_pandas_categorical(dtype):
		return True
	return False

def _is_pandas_categorical(dtype):
	if hasattr(dtype, "name"):
		return dtype.name == "category"
	return False

def _is_ordinal(dtype):
	if _is_pandas_ordinal(dtype):
		return True
	return False

def _is_pandas_ordinal(dtype):
	if isinstance(dtype, CategoricalDtype):
		return dtype.ordered
	return False

class EstimatorProxy(BaseEstimator):

	def __init__(self, estimator, attr_names = ["feature_importances_"]):
		self.estimator = estimator
		self.attr_names = attr_names
		try:
			self._copy_attrs()
		except NotFittedError:
			pass

	def _copy_attrs(self):
		for attr_name in self.attr_names:
			if hasattr(self.estimator, attr_name):
				setattr(self, attr_name, getattr(self.estimator, attr_name))

	def fit(self, X, y = None, **fit_params):
		self.estimator.fit(X, y, **fit_params)
		self._copy_attrs()
		return self

	def predict(self, X, **predict_params):
		return self.estimator.predict(X, **predict_params)

	def predict_proba(self, X, **predict_proba_params):
		return self.estimator.predict_proba(X, **predict_proba_params)

class SelectorProxy(BaseEstimator):

	def __init__(self, selector):
		self.selector = selector
		try:
			self._copy_attrs()
		except NotFittedError:
			pass

	def _copy_attrs(self):
		try:
			setattr(self, "support_mask_", self.selector._get_support_mask())
		except ValueError:
			pass

	def fit(self, X, y = None, **fit_params):
		self.selector.fit(X, y, **fit_params)
		self._copy_attrs()
		return self

	def transform(self, X):
		return self.selector.transform(X)

def _escape(obj, escape_func):
	if isinstance(obj, DataFrameMapper):
		obj.features = _escape_steps(obj.features, escape_func = escape_func)
		if hasattr(obj, "built_features"):
			if obj.built_features is not None:
				obj.built_features = _escape_steps(obj.built_features, escape_func = escape_func)
	elif isinstance(obj, ColumnTransformer):
		obj.transformers = _escape_steps(obj.transformers, escape_func = escape_func)
		obj.remainder = escape_func(obj.remainder, escape_func = escape_func)
		if hasattr(obj, "transformers_"):
			obj.transformers_ = _escape_steps(obj.transformers_, escape_func = escape_func)
	elif isinstance(obj, FeatureUnion):
		obj.transformer_list = _escape_steps(obj.transformer_list, escape_func = escape_func)
	elif isinstance(obj, Pipeline):
		obj.steps = _escape_steps(obj.steps, escape_func = escape_func)
	elif isinstance(obj, SelectorMixin):
		return SelectorProxy(obj)
	elif isinstance(obj, list):
		return [escape_func(e, escape_func = escape_func) for e in obj]
	return obj

def _escape_steps(steps, escape_func):
	return [(step[:1] + (escape_func(step[1], escape_func = escape_func), ) + step[2:]) for step in steps]

def make_pmml_pipeline(estimator, active_fields = None, target_fields = None, escape_func = _escape):
	"""Wraps a Scikit-Learn estimator or pipeline object into a PMML pipeline object.

	Parameters:
	----------
	estimator: BaseEstimator
		The estimator or pipeline object.

	active_fields: list of strings, optional
		Feature names. If missing, "x1", "x2", .., "xn" are assumed.

	target_fields: list of strings, optional
		Label name(s). If missing, "y" is assumed.

	"""
	if not isinstance(estimator, BaseEstimator):
		raise TypeError("The estimator object is not an instance of {0}".format(BaseEstimator.__name__))
	pipeline = PMMLPipeline([
		("estimator", estimator)
	])
	pipeline = _escape(pipeline, escape_func = escape_func)
	if active_fields is not None:
		pipeline.active_fields = numpy.asarray(active_fields)
	if target_fields is not None:
		pipeline.target_fields = numpy.asarray(target_fields)
	return pipeline

def _make_java_command(java_home, java_opts, java_args):
	result = []
	if java_home is not None:
		if not isinstance(java_home, str):
			raise ValueError()
		result.extend([os.path.join(java_home, "bin", "java")])
	else:
		result.extend(["java"])
	if java_opts is not None:
		if not isinstance(java_opts, list):
			raise ValueError()
		result.extend(java_opts)
	if not isinstance(java_args, list):
		raise ValueError()
	result.extend(java_args)
	return result

def _java_version(java_home = None):
	cmd = _make_java_command(java_home = java_home, java_opts = None, java_args = ["-version"])
	try:
		process = Popen(cmd, stdout = PIPE, stderr = PIPE, bufsize = 1, universal_newlines = True)
	except:
		return None
	output, error = process.communicate()
	retcode = process.poll()
	if retcode:
		return None
	return _parse_java_version(error)

def _parse_java_version(java_version):
	match = re.match("^(.*)\sversion\s\"(.*)\"(|\s.+)$", java_version, re.MULTILINE)
	if match:
		return (match.group(1), match.group(2))
	else:
		return None

def _classpath(user_classpath):
	return _package_classpath() + user_classpath

def _process_classpath(name, fun, user_classpath):
	jars = _classpath(user_classpath)
	for jar in jars:
		with ZipFile(jar, "r") as zipfile:
			try:
				zipentry = zipfile.getinfo(name)
			except KeyError:
				pass
			else:
				fun(zipfile.open(zipentry))

def _dill_dump(obj, prefix):
	fd, path = tempfile.mkstemp(prefix = (prefix + "-"), suffix = ".pkl")
	try:
		with open(path, "wb") as dill_file:
			dill.dump(obj, dill_file)
	finally:
		os.close(fd)
	return path

def _joblib_dump(obj, prefix):
	fd, path = tempfile.mkstemp(prefix = (prefix + "-"), suffix = ".pkl.z")
	try:
		joblib.dump(obj, path, compress = 3)
	finally:
		os.close(fd)
	return path

def _is_supported(estimator):
	if hasattr(estimator, "download_mojo"):
		return True
	return isinstance(estimator, BaseEstimator)

def sklearn2pmml(estimator, pmml_path, with_repr = False, java_home = None, java_opts = None, user_classpath = [], dump_flavour = "joblib", debug = False):
	"""Converts a fitted estimator or pipeline object to PMML.

	Parameters:
	----------
	estimator: BaseEstimator
		The estimator or pipeline object.

	pmml_path: string
		The file path where the PMML document should be saved.

	with_repr: boolean, optional
		If true, insert a string containing a printable representation of the estimator object into the PMML document.

	java_home: string, optional
		The path to Java installation directory.
		Functionally analogous to the JAVA_HOME environment variable.

	java_opts: list of strings, optional
		Java options.
		Functionally analogous to the JAVA_OPTS environment variable.

	user_classpath: list of strings, optional
		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
		The SkLearn2PMML classpath is constructed by appending user JAR files to package JAR files.

	dump_flavour: string, optional
		The flavour of pickle dump files.

	debug: boolean, optional
		If true, print information about the conversion process.

	"""
	if debug:
		java_version = _java_version(java_home = java_home)
		if java_version is None:
			java_version = ("java", "N/A")
		print("python: {0}".format(platform.python_version()))
		print("sklearn2pmml: {0}".format(__version__))
		print("sklearn: {0}".format(sklearn.__version__))
		print("sklearn_pandas: {0}".format(sklearn_pandas.__version__))
		print("pandas: {0}".format(pandas.__version__))
		print("numpy: {0}".format(numpy.__version__))
		print("dill: {0}".format(dill.__version__))
		print("joblib: {0}".format(joblib.__version__))
		print("{0}: {1}".format(java_version[0], java_version[1]))
	if not _is_supported(estimator):
		raise TypeError("The estimator object is not an instance of {0}".format(BaseEstimator.__name__))
	# if isinstance(estimator, Pipeline):
	if hasattr(estimator, "_final_estimator"):
		final_estimator = estimator._final_estimator
	else:
		final_estimator = estimator
	dumps = []
	try:
		java_args = ["-cp", os.pathsep.join(_classpath(user_classpath)), "com.sklearn2pmml.Main"]
		if with_repr:
			estimator.repr_ = repr(estimator)
		# if isinstance(final_estimator, H2OEstimator):
		if hasattr(final_estimator, "download_mojo"):
			if dump_flavour != "dill":
				warnings.warn("Changing dump flavour to dill")
				dump_flavour = "dill"
			# Avoid MOJO (re-)download if the indicator attribute is set
			if not (hasattr(final_estimator, "_mojo_path") or hasattr(final_estimator, "_mojo_bytes")):
				mojo_path = final_estimator.download_mojo()
				dumps.append(mojo_path)
				final_estimator._mojo_path = mojo_path
		if dump_flavour == "dill":
			pkl_path = _dill_dump(estimator, "estimator")
		elif dump_flavour == "joblib":
			pkl_path = _joblib_dump(estimator, "estimator")
		else:
			raise ValueError("Dump flavour {0} not in {1}".format(dump_flavour, ["dill", "joblib"]))
		java_args.extend(["--pkl-input", pkl_path])
		dumps.append(pkl_path)
		java_args.extend(["--pmml-output", pmml_path])
		cmd = _make_java_command(java_home = java_home, java_opts = java_opts, java_args = java_args)
		if debug:
			print("Executing command:\n{0}".format(" ".join(cmd)))
		try:
			process = Popen(cmd, stdout = PIPE, stderr = PIPE, bufsize = 1, universal_newlines = True)
		except OSError:
			raise RuntimeError("Java is not installed, or the Java executable is not on system path")
		output, error = process.communicate()
		retcode = process.poll()
		if debug or retcode:
			if len(output) > 0:
				print("Standard output:\n{0}".format(output))
			else:
				print("Standard output is empty")
			if len(error) > 0:
				print("Standard error:\n{0}".format(error))
			else:
				print("Standard error is empty")
		if retcode:
			raise RuntimeError("The SkLearn2PMML application has failed. The Java executable should have printed more information about the failure into its standard output and/or standard error streams")
	finally:
		if debug:
			print("Preserved dump file(s): {0}".format(" ".join(dumps)))
		else:
			for dump in dumps:
				os.remove(dump)

def _parse_properties(lines):
	splitter = re.compile("\s*=\s*")
	properties = dict()
	for line in lines:
		line = line.decode("UTF-8").rstrip()
		if line.startswith("#"):
			continue
		key, value = splitter.split(line)
		properties[key] = value
	return properties

def _format_properties(properties):
	return ["{0} = {1}\n".format(k, v) for k, v in properties.items()]

def _expand_complex_key(key):
	begin = key.find("(")
	end = key.find(")", begin + 1)
	if begin < 0 or end < 0:
		return [key]

	prefix = key[:begin]
	body = key[begin + 1:end]
	suffix = key[end + 1:]

	result = []
	parts = body.split("|")
	for part in parts:
		result += _expand_complex_key(prefix + part + suffix)
	return result

def _expand_mapping(mapping):
	result = dict()
	for k, v in mapping.items():
		pythonClazzes = _expand_complex_key(k)
		javaClazz = v
		for pythonClazz in pythonClazzes:
			result[pythonClazz] = (javaClazz if javaClazz else pythonClazz)
	return result

def load_class_mapping(user_classpath = []):
	"""Loads the class mapping.

	Parameters:
	----------
	user_classpath: list of strings, optional
		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.

	Returns:
	-------
	mapping: dict
		Mapping from Python class names to Java converter class names.

	"""
	mapping = dict()
	processor = lambda x: mapping.update(_expand_mapping(_parse_properties(x.readlines())))
	_process_classpath("META-INF/sklearn2pmml.properties", processor, user_classpath)
	return mapping

def make_class_mapping_jar(mapping, path):
	"""Generates a class mapping JAR file.

	Parameters:
	----------
	mapping: dict of strings
		Mapping from Python class names to Java converter class names.

	path: string
		The path to output JAR file.

	"""
	lines = _format_properties(mapping)

	with ZipFile(path, mode = "w") as zipfile:
		zipfile.writestr("META-INF/sklearn2pmml.properties", "".join(lines))

def _strip_module(name):
	parts = name.split(".")
	if len(parts) > 1:
		parts.pop(-2)
		return ".".join(parts)
	return name



from sklearn2pmml import _strip_module, load_class_mapping

def make_pmml_config(config, user_classpath = []):
	"""Translates a regular TPOT configuration to a PMML-compatible TPOT configuration.

	Parameters:
	----------
	obj: config
		The configuration dictionary.

	user_classpath: list of strings, optional
		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
		The SkLearn2PMML classpath is constructed by appending user JAR files to package JAR files.

	"""
	keys = set(config.keys())
	mapping = load_class_mapping(user_classpath)
	classes = mapping.keys()
	pmml_keys = (set(classes)).union(set([_strip_module(class_) for class_ in classes]))
	return { key : config[key] for key in (keys).intersection(pmml_keys)}


from h2o.estimators import H2OGenericEstimator
from sklearn2pmml import EstimatorProxy

class H2OEstimatorProxy(EstimatorProxy):

	def __init__(self, estimator):
		super(H2OEstimatorProxy, self).__init__(estimator = estimator, attr_names = ["_estimator_type"])

	def download_mojo(self, path = ".", get_genmodel_jar = False, genmodel_name = ""):
		return self.estimator.download_mojo(path = path, get_genmodel_jar = get_genmodel_jar, genmodel_name = genmodel_name)

	def __getstate__(self):
		if not hasattr(self, "_mojo_path"):
			raise AttributeError("The MOJO path is not set")
		state = super(H2OEstimatorProxy, self).__getstate__()
		state["estimator"] = None
		return state

	def __setstate__(self, state):
		state["estimator"] = H2OGenericEstimator.from_file(state["_mojo_path"])
		super(H2OEstimatorProxy, self).__setstate__(state)



from pandas import CategoricalDtype, DataFrame
from pandas.api.types import is_object_dtype
from sklearn.base import clone, BaseEstimator, TransformerMixin
try:
	# SkLearn 1.2.0+
	from sklearn.base import OneToOneFeatureMixin
except:
	from sklearn.base import _OneToOneFeatureMixin as OneToOneFeatureMixin
from sklearn2pmml import _is_pandas_categorical
from sklearn2pmml.util import cast, common_dtype, is_1d, to_numpy

import copy
import itertools
import numpy
import numbers
import pandas

class TransformerWrapper(BaseEstimator, TransformerMixin):

	def __init__(self, transformer, prefit = False):
		self.transformer = transformer
		self.prefit = prefit
		if prefit:
			self.transformer_ = copy.deepcopy(self.transformer)

	def fit(self, X, y = None, **fit_params):
		self.transformer_ = clone(self.transformer)
		if y is None:
			self.transformer_.fit(X, **fit_params)
		else:
			self.transformer_.fit(X, y, **fit_params)
		return self

	def transform(self, X):
		return self.transformer_.transform(X)

class Alias(TransformerWrapper):

	def __init__(self, transformer, name, prefit = False):
		super(Alias, self).__init__(transformer = transformer, prefit = prefit)
		if not isinstance(name, str):
			raise TypeError("Name is not a string")
		self.name = name

	def get_feature_names(self, input_features = None):
		return self.get_feature_names_out(input_features)

	def get_feature_names_out(self, input_features = None):
		return numpy.asarray([self.name])

class MultiAlias(TransformerWrapper):

	def __init__(self, transformer, names, prefit = True):
		super(MultiAlias, self).__init__(transformer = transformer, prefit = prefit)
		for name in names:
			if not isinstance(name, str):
				raise TypeError("Name is not a string")
		self.names = names

	def get_feature_names(self, input_features = None):
		return self.get_feature_names_out(input_features)

	def get_feature_names_out(self, input_features = None):
		return numpy.asarray(self.names)

def _check_cols(X, values):
	if is_1d(X):
		if hasattr(values, "__len__") and len(values) > 1:
			raise ValueError()
	else:
		if X.shape[1] != len(values):
			raise ValueError()

def _set_values(X, where, values):
	X[where] = values
	return X

def _count(missing_mask, valid_mask, invalid_mask):
	missing_freq = sum(missing_mask)
	valid_freq = sum(valid_mask)
	invalid_freq = sum(invalid_mask) if (invalid_mask is not None) else (valid_freq - valid_freq) # A scalar zero, or an array of zeroes
	return {
		"totalFreq" : (missing_freq + valid_freq + invalid_freq),
		"missingFreq" : missing_freq,
		"invalidFreq" : invalid_freq
	}

class Domain(BaseEstimator, TransformerMixin, OneToOneFeatureMixin):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, with_data = True, with_statistics = False, dtype = None, display_name = None):
		self.missing_values = missing_values
		missing_value_treatments = ["as_is", "as_mean", "as_mode", "as_median", "as_value", "return_invalid"]
		if missing_value_treatment not in missing_value_treatments:
			raise ValueError("Missing value treatment {0} not in {1}".format(missing_value_treatment, missing_value_treatments))
		self.missing_value_treatment = missing_value_treatment
		if missing_value_replacement is not None:
			if missing_value_treatment == "return_invalid":
				raise ValueError("Missing value treatment {0} does not support missing_value_replacement attribute".format(missing_value_treatment))
		self.missing_value_replacement = missing_value_replacement
		invalid_value_treatments = ["return_invalid", "as_is", "as_missing", "as_value"]
		if invalid_value_treatment not in invalid_value_treatments:
			raise ValueError("Invalid value treatment {0} not in {1}".format(invalid_value_treatment, invalid_value_treatments))
		self.invalid_value_treatment = invalid_value_treatment
		if invalid_value_replacement is not None:
			if invalid_value_treatment != "as_value":
				raise ValueError("Invalid value treatment {0} does not support invalid_value_replacement attribute".format(invalid_value_treatment))
		else:
			if invalid_value_treatment == "as_value":
				raise ValueError("Invalid value treatment {0} requires invalid_value_replacement attribute".format(invalid_value_treatment))
		self.invalid_value_replacement = invalid_value_replacement
		self.with_data = with_data
		self.with_statistics = with_statistics
		self.dtype = dtype
		self.display_name = display_name

	def _empty_fit(self):
		return not (self.with_data or self.with_statistics)

	def _to_missing(self, X, where):
		if not numpy.any(where):
			return X
		missing_value = None
		if self.missing_values is not None:
			if type(self.missing_values) is list:
				missing_value = self.missing_values[0]
			else:
				missing_value = self.missing_values
		X = _set_values(X, where, missing_value)
		return X

	def _missing_value_mask(self, X):
		if self.missing_values is not None:
			def is_missing(X, missing_value):
				# float("NaN") != float("NaN")
				if isinstance(missing_value, float) and numpy.isnan(missing_value):
					return pandas.isnull(X)
				return X == missing_value

			if type(self.missing_values) is list:
				mask = numpy.full(X.shape, fill_value = False)
				for missing_value in self.missing_values:
					mask = numpy.logical_or(mask, is_missing(X, missing_value))
				return mask
			else:
				return is_missing(X, self.missing_values)
		else:
			return pandas.isnull(X)

	def _valid_value_mask(self, X, where):
		return where

	def _transform_missing_values(self, X, where):
		if not numpy.any(where):
			return X
		if self.missing_value_treatment == "return_invalid":
			raise ValueError("Data contains {0} missing values".format(numpy.count_nonzero(where)))
		elif self.missing_value_treatment in ["as_is", "as_mean", "as_mode", "as_median", "as_value"]:
			if self.missing_value_replacement is not None:
				X = _set_values(X, where, self.missing_value_replacement)
			# Special case for object data type columns: replacing non-None values with None values
			elif is_object_dtype(self.dtype_) and (self.missing_value_replacement is None):
				X = _set_values(X, where, self.missing_value_replacement)
		else:
			raise ValueError()
		return X

	def _transform_valid_values(self, X, where):
		return X

	def _transform_invalid_values(self, X, where):
		if not numpy.any(where):
			return X
		if self.invalid_value_treatment == "return_invalid":
			raise ValueError("Data contains {0} invalid values".format(numpy.count_nonzero(where)))
		elif self.invalid_value_treatment == "as_is":
			pass
		elif self.invalid_value_treatment == "as_missing":
			X = self._to_missing(X, where)
			X = self._transform_missing_values(X, where)
		elif self.invalid_value_treatment == "as_value":
			if self.invalid_value_replacement is not None:
				X = _set_values(X, where, self.invalid_value_replacement)
		else:
			raise ValueError()
		return X

	def _compute_masks(self, X):
		X = to_numpy(X)
		missing_mask = self._missing_value_mask(X)
		nonmissing_mask = ~missing_mask
		valid_mask = self._valid_value_mask(X, nonmissing_mask)
		invalid_mask = ~numpy.logical_or(missing_mask, valid_mask)
		return (missing_mask, valid_mask, invalid_mask)

	def _should_make_copy(self, X, missing_mask, valid_mask, invalid_mask):
		if numpy.any(missing_mask) or numpy.any(invalid_mask):
			return True
		return False

	def transform(self, X):
		if self.dtype is not None:
			X = cast(X, self.dtype)
		missing_mask, valid_mask, invalid_mask = self._compute_masks(X)
		if self._should_make_copy(X, missing_mask, valid_mask, invalid_mask):
			X = X.copy()
		X = self._transform_missing_values(X, missing_mask)
		X = self._transform_valid_values(X, valid_mask)
		X = self._transform_invalid_values(X, invalid_mask)
		return X

class DiscreteDomain(Domain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, with_data = True, with_statistics = False, dtype = None, display_name = None, data_values = None):
		super(DiscreteDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, with_data = with_data, with_statistics = with_statistics, dtype = dtype, display_name = display_name)
		if data_values:
			if not with_data:
				raise ValueError("Valid values require with_data attribute")
			if isinstance(dtype, CategoricalDtype) and data_values != (dtype.categories).tolist():
				raise ValueError("Valid values are invalid")
		self.data_values = data_values

	def _is_ordered(self):
		raise NotImplementedError()

	def _valid_value_mask(self, X, where):
		if hasattr(self, "data_values_"):
			data_values = self.data_values_
		elif _is_pandas_categorical(self.dtype_):
			data_values = self.dtype_.categories
		else:
			data_values = None
		if data_values is not None:
			def _isin_mask(x, values):
				if hasattr(x, "isin"):
					return x.isin(values)
				else:
					return numpy.isin(x, values)

			if is_1d(X):
				where = where.ravel()
				mask = numpy.full(X.shape, fill_value = False)
				mask[where] = _isin_mask(X[where], data_values)
				return mask
			else:
				if hasattr(data_values, "__len__"):
					_check_cols(X, data_values)
				mask = numpy.full(X.shape, fill_value = False)
				for col in range(X.shape[1]):
					col_where = where[:, col]
					col_data_values = data_values[col] if hasattr(data_values, "__len__") else data_values
					mask[col_where, col] = _isin_mask(X[col_where, col], col_data_values)
				return mask
		return super(DiscreteDomain, self)._valid_value_mask(X, where)

	def fit(self, X, y = None):
		if self.dtype is not None:
			if isinstance(self.dtype, str) and self.dtype == "category":
				if self.data_values is not None:
					dtype = CategoricalDtype(list(itertools.chain.from_iterable(self.data_values)), ordered = self._is_ordered())
				else:
					dtype = self.dtype
				X = cast(X, dtype)
			else:
				X = cast(X, self.dtype)
		self.dtype_ = common_dtype(X)
		if self._empty_fit():
			return self
		X = to_numpy(X)
		if self.with_data:
			if self.data_values is None:
				missing_mask = self._missing_value_mask(X)
				nonmissing_mask = ~missing_mask
			else:
				_check_cols(X, self.data_values)
			if is_1d(X):
				if self.data_values is None:
					if _is_pandas_categorical(self.dtype_):
						data_values = self.dtype_.categories
					else:
						data_values = numpy.unique(X[nonmissing_mask])
				else:
					data_values = numpy.asarray(self.data_values)
				self.data_values_ = data_values
			else:
				if self.data_values is None:
					if _is_pandas_categorical(self.dtype_):
						raise ValueError()
				self.data_values_ = []
				for col in range(X.shape[1]):
					if self.data_values is None:
						col_X = X[:, col]
						col_missing_mask = missing_mask[:, col]
						col_nonmissing_mask = nonmissing_mask[:, col]
						data_values = numpy.unique(col_X[col_nonmissing_mask])
					else:
						data_values = numpy.asarray(self.data_values[col])
					self.data_values_.append(data_values)
		if self.with_statistics:
			missing_mask, valid_mask, invalid_mask = self._compute_masks(X)
			self.counts_ = _count(missing_mask, valid_mask, invalid_mask)
			if is_1d(X):
				values, counts = numpy.unique(X[valid_mask], return_counts = True)
				self.discr_stats_ = (values, counts)
			else:
				self.discr_stats_ = []
				for col in range(X.shape[1]):
					col_X = X[:, col]
					col_valid_mask = valid_mask[:, col]
					values, counts = numpy.unique(col_X[col_valid_mask], return_counts = True)
					self.discr_stats_.append((values, counts))
		return self

class CategoricalDomain(DiscreteDomain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, with_data = True, with_statistics = False, dtype = None, display_name = None, data_values = None):
		super(CategoricalDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, with_data = with_data, with_statistics = with_statistics, dtype = dtype, display_name = display_name, data_values = data_values)
		if isinstance(dtype, CategoricalDtype) and dtype.ordered:
			raise ValueError()

	def _is_ordered(self):
		return False

class OrdinalDomain(DiscreteDomain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, with_data = True, with_statistics = False, dtype = None, display_name = None, data_values = None):
		super(OrdinalDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, with_data = with_data, with_statistics = with_statistics, dtype = dtype, display_name = display_name, data_values = data_values)
		if isinstance(dtype, CategoricalDtype) and not dtype.ordered:
			raise ValueError()

	def _is_ordered(self):
		return True

def _interquartile_range(X, axis):
	quartiles = numpy.nanpercentile(X, [25, 75], axis = axis)
	return (quartiles[1] - quartiles[0])

class ContinuousDomain(Domain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, outlier_treatment = "as_is", low_value = None, high_value = None, with_data = True, with_statistics = False, dtype = None, display_name = None, data_min = None, data_max = None):
		super(ContinuousDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, with_data = with_data, with_statistics = with_statistics, dtype = dtype, display_name = display_name)
		outlier_treatments = ["as_is", "as_missing_values", "as_extreme_values"]
		if outlier_treatment not in outlier_treatments:
			raise ValueError("Outlier treatment {0} not in {1}".format(outlier_treatment, outlier_treatments))
		self.outlier_treatment = outlier_treatment
		if outlier_treatment == "as_is":
			if (low_value is not None) or (high_value is not None):
				raise ValueError("Outlier treatment {0} does not support low_value and high_value attributes".format(outlier_treatment))
		elif outlier_treatment == "as_missing_values" or outlier_treatment == "as_extreme_values":
			if (low_value is None) or (high_value is None):
				raise ValueError("Outlier treatment {0} requires low_value and high_value attributes".format(outlier_treatment))
		self.low_value = low_value
		self.high_value = high_value
		if data_min or data_max:
			if not with_data:
				raise ValueError("Valid value intervals require with_data attribute")
		if data_min and data_max:
			if numpy.any(numpy.greater(data_min, data_max)):
				raise ValueError("Valid value intervals are invalid")
		self.data_min = data_min
		self.data_max = data_max

	def _valid_value_mask(self, X, where):
		if hasattr(self, "data_min_") and hasattr(self, "data_max_"):
			return numpy.where((X >= self.data_min_) & (X <= self.data_max_), where, False)
		return super(ContinuousDomain, self)._valid_value_mask(X, where)

	def fit(self, X, y = None):
		if self.dtype is not None:
			X = cast(X, self.dtype)
		self.dtype_ = common_dtype(X)
		if self._empty_fit():
			return self
		X = to_numpy(X)
		if self.with_data:
			if issubclass(self.dtype_.type, numbers.Integral):
				info = numpy.iinfo(self.dtype_)
			else:
				info = numpy.finfo(self.dtype_)
			missing_mask = self._missing_value_mask(X)
			nonmissing_mask = ~missing_mask
			if self.data_min is None:
				self.data_min_ = numpy.asarray(numpy.nanmin(X, axis = 0, initial = info.max, where = nonmissing_mask))
			else:
				_check_cols(X, self.data_min)
				self.data_min_ = numpy.asarray(self.data_min)
			if self.data_max is None:
				self.data_max_ = numpy.asarray(numpy.nanmax(X, axis = 0, initial = info.min, where = nonmissing_mask))
			else:
				_check_cols(X, self.data_max)
				self.data_max_ = numpy.asarray(self.data_max)
		if self.with_statistics:
			missing_mask, valid_mask, invalid_mask = self._compute_masks(X)
			self.counts_ = _count(missing_mask, valid_mask, invalid_mask)
			if numpy.any(missing_mask) or numpy.any(invalid_mask):
				X = (X.copy()).astype(float)
				X[missing_mask | invalid_mask] = float("NaN")
			self.numeric_info_ = {
				"minimum" : numpy.asarray(numpy.nanmin(X, axis = 0)),
				"maximum" : numpy.asarray(numpy.nanmax(X, axis = 0)),
				"mean" : numpy.asarray(numpy.nanmean(X, axis = 0)),
				"standardDeviation" : numpy.asarray(numpy.nanstd(X, axis = 0)),
				"median" : numpy.asarray(numpy.nanmedian(X, axis = 0)),
				"interQuartileRange" : numpy.asarray(_interquartile_range(X, axis = 0))
			}
		return self

	def _outlier_mask(self, X, where):
		return numpy.where((X < self.low_value) | (X > self.high_value), where, False)

	def _negative_outlier_mask(self, X, where):
		return numpy.where(X < self.low_value, where, False)

	def _positive_outlier_mask(self, X, where):
		return numpy.where(X > self.high_value, where, False)

	def _transform_valid_values(self, X, where):
		if not numpy.any(where):
			return X
		if self.outlier_treatment == "as_missing_values":
			outlier_mask = self._outlier_mask(X, where)
			X = self._to_missing(X, outlier_mask)
			X = self._transform_missing_values(X, outlier_mask)
		elif self.outlier_treatment == "as_extreme_values":
			outlier_mask = self._negative_outlier_mask(X, where)
			X = _set_values(X, outlier_mask, self.low_value)
			outlier_mask = self._positive_outlier_mask(X, where)
			X = _set_values(X, outlier_mask, self.high_value)
		return X

	def _should_make_copy(self, X, missing_mask, valid_mask, invalid_mask):
		if super(ContinuousDomain, self)._should_make_copy(X, missing_mask, valid_mask, invalid_mask):
			return True
		if self.outlier_treatment == "as_missing_values":
			outlier_mask = self._outlier_mask(X, valid_mask)
			if numpy.any(outlier_mask):
				return True
		elif self.outlier_treatment == "as_extreme_values":
			outlier_mask = self._negative_outlier_mask(X, valid_mask)
			if numpy.any(outlier_mask):
				return True
			outlier_mask = self._positive_outlier_mask(X, valid_mask)
			if numpy.any(outlier_mask):
				return True
		return False

class TemporalDomain(Domain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, dtype = None, display_name = None):
		super(TemporalDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, with_data = False, with_statistics = False, dtype = dtype, display_name = display_name)
		dtypes = ["datetime64[D]", "datetime64[s]"]
		if (not isinstance(dtype, str)) or (dtype not in dtypes):
			raise ValueError("Temporal data type {0} not in {1}".format(dtype, dtypes))

	def fit(self, X, y = None):
		self.dtype_ = common_dtype(X)
		return self

class DateDomain(TemporalDomain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, display_name = None):
		super(DateDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, dtype = "datetime64[D]", display_name = display_name)

class DateTimeDomain(TemporalDomain):

	def __init__(self, missing_values = None, missing_value_treatment = "as_is", missing_value_replacement = None, invalid_value_treatment = "return_invalid", invalid_value_replacement = None, display_name = None):
		super(DateTimeDomain, self).__init__(missing_values = missing_values, missing_value_treatment = missing_value_treatment, missing_value_replacement = missing_value_replacement, invalid_value_treatment = invalid_value_treatment, invalid_value_replacement = invalid_value_replacement, dtype = "datetime64[s]", display_name = display_name)

class MultiDomain(BaseEstimator, TransformerMixin):

	def __init__(self, domains):
		self.domains = domains

	def fit(self, X, y = None):
		rows, columns = X.shape
		if len(self.domains) != columns:
			raise ValueError("The number of columns {0} is not equal to the number of domain objects {1}".format(columns, len(self.domains)))
		if isinstance(X, DataFrame):
			for domain, column in zip(self.domains, X.columns):
				if domain is not None:
					domain.fit(X[column])
		else:
			for domain, column in zip(self.domains, range(0, columns)):
				if domain is not None:
					domain.fit(X[:, column])
		return self

	def transform(self, X):
		rows, columns = X.shape
		if len(self.domains) != columns:
			raise ValueError("The number of columns {0} is not equal to the number of domain objects {1}".format(columns, len(self.domains)))
		# XXX
		X = X.copy()
		if isinstance(X, DataFrame):
			for domain, column in zip(self.domains, X.columns):
				if domain is not None:
					X[column] = domain.transform(X[column])
		else:
			for domain, column in zip(self.domains, range(0, columns)):
				if domain is not None:
					X[:, column] = domain.transform(X[:, column])
		return X

class DomainEraser(BaseEstimator, TransformerMixin):

	def __init__(self):
		pass

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return X

class ContinuousDomainEraser(DomainEraser):

	def __init__(self):
		super(ContinuousDomainEraser, self).__init__()

class DiscreteDomainEraser(DomainEraser):

	def __init__(self):
		super(DiscreteDomainEraser, self).__init__()